#include <cstdlib>
#include <msp/fs/utils.h>
#include <msp/io/print.h>
#include <msp/strings/lexicalcast.h>
#include <msp/strings/utils.h>
#include "binarypackage.h"
#include "booleanevaluator.h"
#include "builder.h"
#include "file.h"
#include "pkgconfigfile.h"
#include "tool.h"
#include "sourcegenerator.h"
#include "sourcepackage.h"

using namespace std;
using namespace Msp;

namespace {

bool component_sort(const Component &c1, const Component &c2)
{ return c1.get_type()<c2.get_type(); }

}


SourcePackage::SourcePackage(Builder &b, const string &n, const FS::Path &f):
	Package(b, n),
	source_dir(FS::dirname(f)),
	build_type(0),
	config(*this),
	cache(*this)
{
	config.load();

	build_file = builder.get_vfs().get_target(f);
	if(!build_file)
		build_file = new File(builder, *this, f);
	components.push_back(Component(*this, Component::TARBALL, "@src"));
	source_tarball = &components.back();
}

FS::Path SourcePackage::get_temp_directory() const
{
	string subdir = builder.get_current_arch().get_name();
	if(build_type)
	{
		subdir += '.';
		subdir += build_type->get_name();
	}

	const FS::Path &temp = builder.get_temp_directory();
	if(temp.is_absolute())
		return temp/name/subdir;
	else
		return source_dir/temp/subdir;
}

FS::Path SourcePackage::get_output_directory() const
{
	const Architecture &arch = builder.get_current_arch();
	if(arch.is_native())
		return source_dir;
	else
		return source_dir/arch.get_name();
}

bool SourcePackage::match_feature(const string &feat, const string *comp) const
{
	string value = config.get_option("with_"+feat).value;
	if(comp)
		return value==*comp;
	else
		return lexical_cast<bool>(value);
}

void SourcePackage::set_build_type(const BuildType &t)
{
	build_type = &t;
}

void SourcePackage::do_prepare()
{
	BuildInfo final_build_info;

	if(build_type)
		final_build_info.update_from(build_type->get_build_info());

	final_build_info.update_from(build_info);
	build_info = final_build_info;

	build_info.incpath.push_back((builder.get_prefix()/"include").str());
	build_info.libpath.push_back((builder.get_prefix()/"lib").str());

	for(FeatureList::iterator i=features.begin(); i!=features.end(); ++i)
	{
		string ident = "WITH_"+toupper(i->name);
		string value = config.get_option("with_"+i->name).value;
		if(!i->choices.empty())
			build_info.defines[ident] = value;
		else if(lexical_cast<bool>(value))
			build_info.defines[ident] = "1";
	}

	bool export_paths = false;
	for(list<Component>::iterator i=components.begin(); i!=components.end(); ++i)
	{
		i->prepare();
		i->create_build_info();

		if(i->get_type()==Component::LIBRARY)
		{
			export_binfo.libs.push_back(i->get_name());
			export_paths = true;
		}
	}

	if(export_paths)
	{
		export_binfo.incpath.push_back((builder.get_prefix()/"include").str());
		export_binfo.libpath.push_back((builder.get_prefix()/"lib").str());
	}

	cache.load();

	bool pc_needed = false;
	for(ComponentList::const_iterator i=components.begin(); i!=components.end(); ++i)
	{
		i->create_targets();
		if(i->get_type()==Component::LIBRARY)
			pc_needed = true;
	}

	if(pc_needed)
	{
		PkgConfigFile *pc = new PkgConfigFile(builder, *this);
		builder.get_build_graph().get_target("install")->add_dependency(*builder.get_toolchain().get_tool("CP").create_target(*pc));
	}
}

void SourcePackage::save_caches()
{
	config.save();
	cache.save();
}


SourcePackage::Loader::Loader(SourcePackage &p):
	DataFile::DerivedObjectLoader<SourcePackage, Package::Loader>(p)
{
	init(0);
}

SourcePackage::Loader::Loader(SourcePackage &p, const Config::InputOptions &o):
	DataFile::DerivedObjectLoader<SourcePackage, Package::Loader>(p)
{
	init(&o);
}

void SourcePackage::Loader::init(const Config::InputOptions *o)
{
	options = o;
	add("description", &SourcePackage::description);
	add("build_info",  &Loader::build_info);
	add("feature",     &Loader::feature);
	add("generate",    &Loader::generate);
	add("if_feature",  &Loader::if_feature);
	add("program",     &Loader::component<Component::PROGRAM>);
	add("library",     &Loader::component<Component::LIBRARY>);
	add("module",      &Loader::component<Component::MODULE>);
	add("install",     &Loader::component<Component::INSTALL>);
	add("interface_version", &Loader::interface_version);
	add("datapack",    &Loader::component<Component::DATAPACK>);
	add("source_tarball", &Loader::source_tarball);
	add("tarball",     &Loader::tarball);
	add("version",     &Loader::version);
}

void SourcePackage::Loader::finish()
{
	obj.components.sort(component_sort);
}

void SourcePackage::Loader::feature(const string &n, const string &d)
{
	Feature feat(n);
	feat.description = d;
	load_sub(feat);
	obj.features.push_back(feat);

	const Config::Option &opt = obj.config.add_option(feat);
	if(options)
	{
		Config::InputOptions::const_iterator i = options->find(opt.name);
		if(i!=options->end())
			obj.config.set_option(opt.name, i->second);
	}
}

template<Component::Type t>
void SourcePackage::Loader::component(const string &n)
{
	Component comp(obj, t, n);
	load_sub(comp);
	obj.components.push_back(comp);
}

void SourcePackage::Loader::build_info()
{
	load_sub(obj.build_info);
}

void SourcePackage::Loader::generate(const string &tag)
{
	SourceGenerator *gen = new SourceGenerator(obj.builder, obj, tag);
	load_sub(*gen);
	obj.local_tools.add_tool(gen);
}

void SourcePackage::Loader::if_feature(const string &cond)
{
	BooleanEvaluator eval(sigc::mem_fun(&obj, &SourcePackage::match_feature));
	bool match = eval.evaluate(cond);
	obj.builder.get_logger().log("configure", format("%s: feature %s %smatched", obj.name, cond, (match ? "" : "not ")));
	if(match)
		load_sub_with(*this);
}

void SourcePackage::Loader::interface_version(const string &v)
{
	obj.interface_version = v;
	if(obj.version.empty())
		obj.version = v;
}

void SourcePackage::Loader::source_tarball()
{
	load_sub(*obj.source_tarball);
}

void SourcePackage::Loader::tarball(const string &n)
{
	Component trbl(obj, Component::TARBALL, n);
	load_sub(trbl);
}

void SourcePackage::Loader::version(const string &v)
{
	obj.version = v;

	string::size_type i = 0;
	for(unsigned dots=0; i<obj.version.size(); ++i)
		if(obj.version[i]=='.' && ++dots>=2)
			break;
	obj.interface_version = obj.version.substr(0, i);
}
